@@ -41,9 +41,8 @@ OBSAdvAudioCtrl::OBSAdvAudioCtrl(QGridLayout *, obs_source_t *source_)
 	percent = new QSpinBox();
 	forceMono = new QCheckBox();
 	balance = new BalanceSlider();
-#if defined(_WIN32) || defined(__APPLE__) || HAVE_PULSEAUDIO
-	monitoringType = new QComboBox();
-#endif
+	if (obs_audio_monitoring_available())
+		monitoringType = new QComboBox();
 	syncOffset = new QSpinBox();
 	mixer1 = new QCheckBox();
 	mixer2 = new QCheckBox();
@@ -61,8 +60,14 @@ OBSAdvAudioCtrl::OBSAdvAudioCtrl(QGridLayout *, obs_source_t *source_)
 				 this);
 	flagsSignal.Connect(handler, "update_flags", OBSSourceFlagsChanged,
 			    this);
+	if (obs_audio_monitoring_available())
+		monitoringTypeSignal.Connect(handler, "audio_monitoring",
+					     OBSSourceMonitoringTypeChanged,
+					     this);
 	mixersSignal.Connect(handler, "audio_mixers", OBSSourceMixersChanged,
 			     this);
+	balChangedSignal.Connect(handler, "audio_balance",
+				 OBSSourceBalanceChanged, this);
 
 	hlayout = new QHBoxLayout();
 	hlayout->setContentsMargins(0, 0, 0, 0);
@@ -170,19 +175,22 @@ OBSAdvAudioCtrl::OBSAdvAudioCtrl(QGridLayout *, obs_source_t *source_)
 		QTStr("Basic.AdvAudio.SyncOffsetSource").arg(sourceName));
 
 	int idx;
-#if defined(_WIN32) || defined(__APPLE__) || HAVE_PULSEAUDIO
-	monitoringType->addItem(QTStr("Basic.AdvAudio.Monitoring.None"),
-				(int)OBS_MONITORING_TYPE_NONE);
-	monitoringType->addItem(QTStr("Basic.AdvAudio.Monitoring.MonitorOnly"),
-				(int)OBS_MONITORING_TYPE_MONITOR_ONLY);
-	monitoringType->addItem(QTStr("Basic.AdvAudio.Monitoring.Both"),
-				(int)OBS_MONITORING_TYPE_MONITOR_AND_OUTPUT);
-	int mt = (int)obs_source_get_monitoring_type(source);
-	idx = monitoringType->findData(mt);
-	monitoringType->setCurrentIndex(idx);
-	monitoringType->setAccessibleName(
-		QTStr("Basic.AdvAudio.MonitoringSource").arg(sourceName));
-#endif
+	if (obs_audio_monitoring_available()) {
+		monitoringType->addItem(QTStr("Basic.AdvAudio.Monitoring.None"),
+					(int)OBS_MONITORING_TYPE_NONE);
+		monitoringType->addItem(
+			QTStr("Basic.AdvAudio.Monitoring.MonitorOnly"),
+			(int)OBS_MONITORING_TYPE_MONITOR_ONLY);
+		monitoringType->addItem(
+			QTStr("Basic.AdvAudio.Monitoring.Both"),
+			(int)OBS_MONITORING_TYPE_MONITOR_AND_OUTPUT);
+		int mt = (int)obs_source_get_monitoring_type(source);
+		idx = monitoringType->findData(mt);
+		monitoringType->setCurrentIndex(idx);
+		monitoringType->setAccessibleName(
+			QTStr("Basic.AdvAudio.MonitoringSource")
+				.arg(sourceName));
+	}
 
 	mixer1->setText("1");
 	mixer1->setChecked(mixers & (1 << 0));
@@ -237,10 +245,10 @@ OBSAdvAudioCtrl::OBSAdvAudioCtrl(QGridLayout *, obs_source_t *source_)
 			 SLOT(ResetBalance()));
 	QWidget::connect(syncOffset, SIGNAL(valueChanged(int)), this,
 			 SLOT(syncOffsetChanged(int)));
-#if defined(_WIN32) || defined(__APPLE__) || HAVE_PULSEAUDIO
-	QWidget::connect(monitoringType, SIGNAL(currentIndexChanged(int)), this,
-			 SLOT(monitoringTypeChanged(int)));
-#endif
+	if (obs_audio_monitoring_available())
+		QWidget::connect(monitoringType,
+				 SIGNAL(currentIndexChanged(int)), this,
+				 SLOT(monitoringTypeChanged(int)));
 	QWidget::connect(mixer1, SIGNAL(clicked(bool)), this,
 			 SLOT(mixer1Changed(bool)));
 	QWidget::connect(mixer2, SIGNAL(clicked(bool)), this,
@@ -266,9 +274,8 @@ OBSAdvAudioCtrl::~OBSAdvAudioCtrl()
 	forceMonoContainer->deleteLater();
 	balanceContainer->deleteLater();
 	syncOffset->deleteLater();
-#if defined(_WIN32) || defined(__APPLE__) || HAVE_PULSEAUDIO
-	monitoringType->deleteLater();
-#endif
+	if (obs_audio_monitoring_available())
+		monitoringType->deleteLater();
 	mixerContainer->deleteLater();
 }
 
@@ -284,9 +291,8 @@ void OBSAdvAudioCtrl::ShowAudioControl(QGridLayout *layout)
 	layout->addWidget(forceMonoContainer, lastRow, idx++);
 	layout->addWidget(balanceContainer, lastRow, idx++);
 	layout->addWidget(syncOffset, lastRow, idx++);
-#if defined(_WIN32) || defined(__APPLE__) || HAVE_PULSEAUDIO
-	layout->addWidget(monitoringType, lastRow, idx++);
-#endif
+	if (obs_audio_monitoring_available())
+		layout->addWidget(monitoringType, lastRow, idx++);
 	layout->addWidget(mixerContainer, lastRow, idx++);
 	layout->layout()->setAlignment(mixerContainer, Qt::AlignVCenter);
 	layout->setHorizontalSpacing(15);
@@ -330,6 +336,15 @@ void OBSAdvAudioCtrl::OBSSourceSyncChanged(void *param, calldata_t *calldata)
 				  "SourceSyncChanged", Q_ARG(int64_t, offset));
 }
 
+void OBSAdvAudioCtrl::OBSSourceMonitoringTypeChanged(void *param,
+						     calldata_t *calldata)
+{
+	int type = calldata_int(calldata, "type");
+	QMetaObject::invokeMethod(reinterpret_cast<OBSAdvAudioCtrl *>(param),
+				  "SourceMonitoringTypeChanged",
+				  Q_ARG(int, type));
+}
+
 void OBSAdvAudioCtrl::OBSSourceMixersChanged(void *param, calldata_t *calldata)
 {
 	uint32_t mixers = (uint32_t)calldata_int(calldata, "mixers");
@@ -338,6 +353,13 @@ void OBSAdvAudioCtrl::OBSSourceMixersChanged(void *param, calldata_t *calldata)
 				  Q_ARG(uint32_t, mixers));
 }
 
+void OBSAdvAudioCtrl::OBSSourceBalanceChanged(void *param, calldata_t *calldata)
+{
+	int balance = (float)calldata_float(calldata, "balance") * 100.0f;
+	QMetaObject::invokeMethod(reinterpret_cast<OBSAdvAudioCtrl *>(param),
+				  "SourceBalanceChanged", Q_ARG(int, balance));
+}
+
 /* ------------------------------------------------------------------------- */
 /* Qt event queue source callbacks */
 
@@ -375,9 +397,26 @@ void OBSAdvAudioCtrl::SourceVolumeChanged(float value)
 	volume->blockSignals(false);
 }
 
+void OBSAdvAudioCtrl::SourceBalanceChanged(int value)
+{
+	balance->blockSignals(true);
+	balance->setValue(value);
+	balance->blockSignals(false);
+}
+
 void OBSAdvAudioCtrl::SourceSyncChanged(int64_t offset)
 {
+	syncOffset->blockSignals(true);
 	syncOffset->setValue(offset / NSEC_PER_MSEC);
+	syncOffset->blockSignals(false);
+}
+
+void OBSAdvAudioCtrl::SourceMonitoringTypeChanged(int type)
+{
+	int idx = monitoringType->findData(type);
+	monitoringType->blockSignals(true);
+	monitoringType->setCurrentIndex(idx);
+	monitoringType->blockSignals(false);
 }
 
 void OBSAdvAudioCtrl::SourceMixersChanged(uint32_t mixers)
@@ -395,6 +434,8 @@ void OBSAdvAudioCtrl::SourceMixersChanged(uint32_t mixers)
 
 void OBSAdvAudioCtrl::volumeChanged(double db)
 {
+	float prev = obs_source_get_volume(source);
+
 	if (db < MIN_DB) {
 		volume->setSpecialValueText("-inf dB");
 		db = -INFINITY;
@@ -402,30 +443,87 @@ void OBSAdvAudioCtrl::volumeChanged(double db)
 
 	float val = obs_db_to_mul(db);
 	obs_source_set_volume(source, val);
+
+	auto undo_redo = [](const std::string &name, float val) {
+		OBSSourceAutoRelease source =
+			obs_get_source_by_name(name.c_str());
+		obs_source_set_volume(source, val);
+	};
+
+	const char *name = obs_source_get_name(source);
+
+	OBSBasic *main = OBSBasic::Get();
+	main->undo_s.add_action(
+		QTStr("Undo.Volume.Change").arg(name),
+		std::bind(undo_redo, std::placeholders::_1, prev),
+		std::bind(undo_redo, std::placeholders::_1, val), name, name,
+		true);
 }
 
 void OBSAdvAudioCtrl::percentChanged(int percent)
 {
-	obs_source_set_volume(source, (float)percent / 100.0f);
+	float prev = obs_source_get_volume(source);
+	float val = (float)percent / 100.0f;
+
+	obs_source_set_volume(source, val);
+
+	auto undo_redo = [](const std::string &name, float val) {
+		OBSSourceAutoRelease source =
+			obs_get_source_by_name(name.c_str());
+		obs_source_set_volume(source, val);
+	};
+
+	const char *name = obs_source_get_name(source);
+	OBSBasic::Get()->undo_s.add_action(
+		QTStr("Undo.Volume.Change").arg(name),
+		std::bind(undo_redo, std::placeholders::_1, prev),
+		std::bind(undo_redo, std::placeholders::_1, val), name, name,
+		true);
 }
 
-void OBSAdvAudioCtrl::downmixMonoChanged(bool checked)
+static inline void set_mono(obs_source_t *source, bool mono)
+{
+	uint32_t flags = obs_source_get_flags(source);
+	if (mono)
+		flags |= OBS_SOURCE_FLAG_FORCE_MONO;
+	else
+		flags &= ~OBS_SOURCE_FLAG_FORCE_MONO;
+	obs_source_set_flags(source, flags);
+}
+
+void OBSAdvAudioCtrl::downmixMonoChanged(bool val)
 {
 	uint32_t flags = obs_source_get_flags(source);
 	bool forceMonoActive = (flags & OBS_SOURCE_FLAG_FORCE_MONO) != 0;
 
-	if (forceMonoActive != checked) {
-		if (checked)
-			flags |= OBS_SOURCE_FLAG_FORCE_MONO;
-		else
-			flags &= ~OBS_SOURCE_FLAG_FORCE_MONO;
+	if (forceMonoActive == val)
+		return;
 
-		obs_source_set_flags(source, flags);
-	}
+	if (val)
+		flags |= OBS_SOURCE_FLAG_FORCE_MONO;
+	else
+		flags &= ~OBS_SOURCE_FLAG_FORCE_MONO;
+
+	obs_source_set_flags(source, flags);
+
+	auto undo_redo = [](const std::string &name, bool val) {
+		OBSSourceAutoRelease source =
+			obs_get_source_by_name(name.c_str());
+		set_mono(source, val);
+	};
+
+	QString text = QTStr(val ? "Undo.ForceMono.On" : "Undo.ForceMono.Off");
+
+	const char *name = obs_source_get_name(source);
+	OBSBasic::Get()->undo_s.add_action(
+		text.arg(name),
+		std::bind(undo_redo, std::placeholders::_1, !val),
+		std::bind(undo_redo, std::placeholders::_1, val), name, name);
 }
 
 void OBSAdvAudioCtrl::balanceChanged(int val)
 {
+	float prev = obs_source_get_balance_value(source);
 	float bal = (float)val / 100.0f;
 
 	if (abs(50 - val) < 10) {
@@ -436,6 +534,19 @@ void OBSAdvAudioCtrl::balanceChanged(int val)
 	}
 
 	obs_source_set_balance_value(source, bal);
+
+	auto undo_redo = [](const std::string &name, float val) {
+		OBSSourceAutoRelease source =
+			obs_get_source_by_name(name.c_str());
+		obs_source_set_balance_value(source, val);
+	};
+
+	const char *name = obs_source_get_name(source);
+	OBSBasic::Get()->undo_s.add_action(
+		QTStr("Undo.Balance.Change").arg(name),
+		std::bind(undo_redo, std::placeholders::_1, prev),
+		std::bind(undo_redo, std::placeholders::_1, bal), name, name,
+		true);
 }
 
 void OBSAdvAudioCtrl::ResetBalance()
@@ -445,17 +556,35 @@ void OBSAdvAudioCtrl::ResetBalance()
 
 void OBSAdvAudioCtrl::syncOffsetChanged(int milliseconds)
 {
-	int64_t cur_val = obs_source_get_sync_offset(source);
+	int64_t prev = obs_source_get_sync_offset(source);
+	int64_t val = int64_t(milliseconds) * NSEC_PER_MSEC;
 
-	if (cur_val / NSEC_PER_MSEC != milliseconds)
-		obs_source_set_sync_offset(source, int64_t(milliseconds) *
-							   NSEC_PER_MSEC);
+	if (prev / NSEC_PER_MSEC == milliseconds)
+		return;
+
+	obs_source_set_sync_offset(source, val);
+
+	auto undo_redo = [](const std::string &name, int64_t val) {
+		OBSSourceAutoRelease source =
+			obs_get_source_by_name(name.c_str());
+		obs_source_set_sync_offset(source, val);
+	};
+
+	const char *name = obs_source_get_name(source);
+	OBSBasic::Get()->undo_s.add_action(
+		QTStr("Undo.SyncOffset.Change").arg(name),
+		std::bind(undo_redo, std::placeholders::_1, prev),
+		std::bind(undo_redo, std::placeholders::_1, val), name, name,
+		true);
 }
 
 void OBSAdvAudioCtrl::monitoringTypeChanged(int index)
 {
-	int mt = monitoringType->itemData(index).toInt();
-	obs_source_set_monitoring_type(source, (obs_monitoring_type)mt);
+	obs_monitoring_type prev = obs_source_get_monitoring_type(source);
+
+	obs_monitoring_type mt =
+		(obs_monitoring_type)monitoringType->itemData(index).toInt();
+	obs_source_set_monitoring_type(source, mt);
 
 	const char *type = nullptr;
 
@@ -471,8 +600,20 @@ void OBSAdvAudioCtrl::monitoringTypeChanged(int index)
 		break;
 	}
 
+	const char *name = obs_source_get_name(source);
 	blog(LOG_INFO, "User changed audio monitoring for source '%s' to: %s",
-	     obs_source_get_name(source), type);
+	     name ? name : "(null)", type);
+
+	auto undo_redo = [](const std::string &name, obs_monitoring_type val) {
+		OBSSourceAutoRelease source =
+			obs_get_source_by_name(name.c_str());
+		obs_source_set_monitoring_type(source, val);
+	};
+
+	OBSBasic::Get()->undo_s.add_action(
+		QTStr("Undo.MonitoringType.Change").arg(name),
+		std::bind(undo_redo, std::placeholders::_1, prev),
+		std::bind(undo_redo, std::placeholders::_1, mt), name, name);
 }
 
 static inline void setMixer(obs_source_t *source, const int mixerIdx,
@@ -487,6 +628,19 @@ static inline void setMixer(obs_source_t *source, const int mixerIdx,
 		new_mixers &= ~(1 << mixerIdx);
 
 	obs_source_set_audio_mixers(source, new_mixers);
+
+	auto undo_redo = [](const std::string &name, uint32_t mixers) {
+		OBSSourceAutoRelease source =
+			obs_get_source_by_name(name.c_str());
+		obs_source_set_audio_mixers(source, mixers);
+	};
+
+	const char *name = obs_source_get_name(source);
+	OBSBasic::Get()->undo_s.add_action(
+		QTStr("Undo.Mixers.Change").arg(name),
+		std::bind(undo_redo, std::placeholders::_1, mixers),
+		std::bind(undo_redo, std::placeholders::_1, new_mixers), name,
+		name);
 }
 
 void OBSAdvAudioCtrl::mixer1Changed(bool checked)
